precision highp float;

varying vec2 textureCoordinate;
uniform sampler2D inputImageTexture;
uniform lowp mat4 colorMatrix;
//uniform mediump vec2 textureSize;
uniform lowp float uImageWidth;
uniform lowp float uImageHeight;

uniform highp float exposure;
uniform lowp float contrast;
uniform lowp float saturation;
uniform lowp float temperature;
uniform lowp float tint;
uniform lowp vec2 vignetteCenter;
uniform lowp vec3 vignetteColor;
uniform highp float vignetteStart;
uniform highp float vignetteEnd;
uniform lowp float shadows;
uniform lowp float highlights;
uniform lowp float ambiance;
uniform lowp float brightness;
uniform lowp float grain;
uniform lowp float fade;
uniform lowp float hueAdjust;
uniform lowp float sharpness;

const mediump vec3 luminanceWeightingHighlight = vec3(0.3, 0.3, 0.3);
const mediump vec3 luminanceWeighting = vec3(0.2125, 0.7154, 0.0721);

const lowp vec3 warmFilter = vec3(0.93, 0.54, 0.0);
const mediump mat3 RGBtoYIQ = mat3(0.299, 0.587, 0.114, 0.596, -0.274, -0.322, 0.212, -0.523, 0.311);
const mediump mat3 YIQtoRGB = mat3(1.0, 0.956, 0.621, 1.0, -0.272, -0.647, 1.0, -1.105, 1.702);

const highp vec4 kRGBToYPrime = vec4 (0.299, 0.587, 0.114, 0.0);
const highp vec4 kRGBToI = vec4 (0.595716, -0.274453, -0.321263, 0.0);
const highp vec4 kRGBToQ = vec4 (0.211456, -0.522591, 0.31135, 0.0);

const highp vec4 kYIQToR = vec4 (1.0, 0.9563, 0.6210, 0.0);
const highp vec4 kYIQToG = vec4 (1.0, -0.2721, -0.6474, 0.0);
const highp vec4 kYIQToB = vec4 (1.0, -1.1070, 1.7046, 0.0);

vec4 adjustHue(vec4 srcColor) {
    // Convert to YIQ
    highp float YPrime = dot (srcColor, kRGBToYPrime);
    highp float I = dot (srcColor, kRGBToI);
    highp float Q = dot (srcColor, kRGBToQ);

    // Calculate the hue and chroma
    highp float hue = atan (Q, I);
    highp float chroma = sqrt (I * I + Q * Q);

    // Make the user's adjustments
    hue += (-hueAdjust);//why negative rotation?

    // Convert back to YIQ
    Q = chroma * sin (hue);
    I = chroma * cos (hue);

    // Convert back to RGB
    highp vec4 yIQ = vec4 (YPrime, I, Q, 0.0);
    srcColor.r = dot (yIQ, kYIQToR);
    srcColor.g = dot (yIQ, kYIQToG);
    srcColor.b = dot (yIQ, kYIQToB);

    return srcColor;
}

// 这个算法是GPUImage框架中的锐化算法
vec4 adjustSharpen(vec4 srcColor) {
    if (uImageWidth <= 0.0 || uImageHeight <= 0.0) {
        // 如果没传宽高，直接不处理
        return srcColor;
    }

    highp vec2 uv = textureCoordinate;

    highp vec2 widthStep = vec2(uImageWidth, 0.0);
    highp vec2 heightStep = vec2(0.0, uImageHeight);

    highp vec2 leftTextureCoordinate = uv - widthStep;
    highp vec2 rightTextureCoordinate = uv + widthStep;
    highp vec2 topTextureCoordinate = uv + heightStep;
    highp vec2 bottomTextureCoordinate = uv - heightStep;

    highp float centerMultiplier = 1.0 + 4.0 * sharpness;
    highp float edgeMultiplier = sharpness;

    highp vec3 leftTextureColor = texture2D(inputImageTexture, leftTextureCoordinate).rgb;
    highp vec3 rightTextureColor = texture2D(inputImageTexture, rightTextureCoordinate).rgb;
    highp vec3 topTextureColor = texture2D(inputImageTexture, topTextureCoordinate).rgb;
    highp vec4 bottomTextureColor = texture2D(inputImageTexture, bottomTextureCoordinate);

    highp vec4 sharpenColor = vec4((srcColor.rgb * centerMultiplier - (leftTextureColor * edgeMultiplier + rightTextureColor * edgeMultiplier + topTextureColor * edgeMultiplier + bottomTextureColor.rgb * edgeMultiplier)), bottomTextureColor.w);

    return sharpenColor;
}

void main() {
    highp vec2 uv = textureCoordinate;
    highp vec4 normalColor = texture2D(inputImageTexture, fract(uv));

    //阴影
    mediump float luminance2 = dot(normalColor.rgb, luminanceWeightingHighlight);
    mediump float shadow;
    if (shadows > 0.0) {
        shadow = clamp((pow(luminance2, 1.0/(shadows+1.0)) + (-0.76)*pow(luminance2, 2.0/(shadows+1.0))) - luminance2, 0.0, 1.0);
    } else {
        shadow = clamp((pow(luminance2, 1.0-shadows) + (0.76)*pow(luminance2, 2.0*(1.0-shadows))) - luminance2, -1.0, 0.0);
    }

    //高光
    mediump float highlight;
    if (highlights < 0.0) {
        highlight = clamp((1.0 - (pow(1.0-luminance2, 1.0/(1.0-highlights)) + (-0.8)*pow(1.0-luminance2, 2.0/(1.0-highlights)))) - luminance2, -1.0, 0.0);
    } else {
        highlight = clamp((1.0 - pow(1.0-luminance2, 1.0+highlights) + (-0.8)*pow(1.0-luminance2, 2.0*(1.0+highlights))) - luminance2, 0.0, 1.0);
    }
    lowp vec3 rs = vec3(0.0, 0.0, 0.0) + ((luminance2 + shadow + highlight) - 0.0) * ((normalColor.rgb - vec3(0.0, 0.0, 0.0))/(luminance2 - 0.0));
    lowp vec4 color = vec4(rs, normalColor.a);

    // 锐化sharpen
    mediump vec4 sharpenColor = adjustSharpen(color);

    //颗粒
    highp float x = (uv.x + 4.0) * (uv.y + 4.0) * 10.0;
    highp vec4 grain = vec4(mod((mod(x, 13.0) + 1.0) * (mod(x, 123.0) + 1.0), 0.01) - 0.005) * grain;
    highp vec4 result =  sharpenColor + grain;

    //曝光度
    lowp vec4 exposureColor = vec4(result.rgb * pow(2.0, exposure), result.w);
    lowp vec4 outputColor = exposureColor * colorMatrix;

    //褪色
    lowp vec4 fadeColor = (fade * outputColor) + ((1.0 - fade) * exposureColor);

    // 色相Hue
    lowp vec4 hueColor = adjustHue(fadeColor);

    //对比度
    lowp vec4 contrastColor = vec4(((hueColor.rgb - vec3(0.5)) * contrast + vec3(0.5)), hueColor.w);

    //饱和度
    lowp float luminance = dot(contrastColor.rgb, luminanceWeighting);
    lowp vec3 greyScaleColor = vec3(luminance);
    lowp vec4 saturationColor = vec4(mix(greyScaleColor, contrastColor.rgb, saturation), contrastColor.w);

    //温度
    mediump vec3 yiq = RGBtoYIQ * saturationColor.rgb;
    yiq.b = clamp(yiq.b + tint*0.5226*0.1, -0.5226, 0.5226);
    lowp vec3 rgb = YIQtoRGB * yiq;
    lowp vec3 processed = vec3(
    (rgb.r < 0.5 ? (2.0 * rgb.r * warmFilter.r) : (1.0 - 2.0 * (1.0 - rgb.r) * (1.0 - warmFilter.r))),
    (rgb.g < 0.5 ? (2.0 * rgb.g * warmFilter.g) : (1.0 - 2.0 * (1.0 - rgb.g) * (1.0 - warmFilter.g))),
    (rgb.b < 0.5 ? (2.0 * rgb.b * warmFilter.b) : (1.0 - 2.0 * (1.0 - rgb.b) * (1.0 - warmFilter.b))));
    lowp vec4 seWenSeDiao = vec4(mix(rgb, processed, temperature), saturationColor.a);

    // vignette
    lowp float d = distance(textureCoordinate, vec2(vignetteCenter.x, vignetteCenter.y));
    lowp float percent = smoothstep(vignetteStart, vignetteEnd, d);
    lowp vec4 vignette = vec4(mix(seWenSeDiao.x, vignetteColor.x, percent), mix(seWenSeDiao.y, vignetteColor.y, percent), mix(seWenSeDiao.z, vignetteColor.z, percent), seWenSeDiao.a);

    //氛围
    lowp vec4 rsShadow = vec4(vignette.rgb, vignette.a);
    lowp float luminance3 = dot(rsShadow.rgb, luminanceWeighting);
    lowp vec3 greyScaleColorFenWei = vec3(luminance3);
    lowp vec4 fenweiSaturation;
    if (ambiance < 0.0) {
        fenweiSaturation = vec4(mix(greyScaleColorFenWei, rsShadow.rgb, abs(ambiance / 5.0) + 1.0), rsShadow.a);
    } else {
        fenweiSaturation = vec4(mix(greyScaleColorFenWei, rsShadow.rgb, ambiance * 3.0 + 1.0), rsShadow.a);
    }
    highp float lum = fenweiSaturation.r * 0.3 + fenweiSaturation.g * 0.59 + fenweiSaturation.b * 0.11;
    lowp vec4 fenweiRs = vec4(fenweiSaturation.r - fenweiSaturation.r * ambiance * (lum - 0.5),
    fenweiSaturation.g - fenweiSaturation.g * ambiance * (lum - 0.5),
    fenweiSaturation.b - fenweiSaturation.b * ambiance * (lum - 0.5),
    fenweiSaturation.a);

    //亮度
    gl_FragColor = vec4((fenweiRs.rgb + vec3(brightness)), fenweiRs.w);
}